#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#define VPCALC_SECTION	"LIB"
#define VPCALC_SUBSECT	"CALC"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/conf.h"

#include "../include/timing.h"

#include "../include/mem.h"
#include "../include/gfx.h"
#include "../include/dev.h"
#include "../include/cpu.h"

#include "../include/calc.h"

typedef struct _calc_o {
	const char *mem_libname;
	const char *gfx_libname;
	const char *dev_libname;
	const char *cpu_libname;

	timing_o timings;

	mem_o mem;
	gfx_o gfx;
	dev_o dev;
	cpu_o cpu;
} _calc_o;

calc_o calc_new(config_o conf, timing_func callback)
{
	_calc_o *new;
	uint depth = 32;
	uint scale = 1;

	new = (_calc_o *) malloc(sizeof(_calc_o));
	if (new == NULL) {
		_err_report("Unable to allocate calc structure.");

		return NULL;
	}

	new->mem_libname = config_string(conf, "mem_driver");
	new->gfx_libname = config_string(conf, "gfx_driver");
	new->dev_libname = config_string(conf, "dev_driver");
	new->cpu_libname = config_string(conf, "cpu_driver");

	if (config_isset(conf, "gfx_depth"))
		depth = config_integer(conf, "gfx_depth");
	if (config_isset(conf, "gfx_scale"))
		scale = config_integer(conf, "gfx_scale");

	new->timings.cpint = config_integer(conf, "cycles_per_interrupt");
	if(new->timings.cpint <= 0) {
		_err_report("Bad value for cycles per interrupt!");

		return NULL;
	}

	new->timings.intps = config_integer(conf, "interrupts_per_second");
	if(new->timings.intps <= 0) {
		_err_report("Bad value for interrupts per second!");

		return NULL;
	}

	new->timings.tick_rate = 1000 / new->timings.intps;
	new->timings.ticks = new->timings.tick_rate;
	new->timings.error = 1000;
	new->timings.overflow = 0;
	new->timings.callback = callback;

	new->mem = mem_new(new->mem_libname);
	if (new->mem == NULL) {
		free(new);
		return NULL;
	}

	new->gfx = gfx_new(new->gfx_libname, new->mem, depth, scale);
	if (new->gfx == NULL) {
		mem_del(new->mem);
		free(new);
		return NULL;
	}

	new->dev = dev_new(new->dev_libname, new->mem, new->gfx);
	if (new->dev == NULL) {
		gfx_del(new->gfx);
		mem_del(new->mem);
		free(new);
		return NULL;
	}

	new->cpu = cpu_new(new->cpu_libname, new->mem, new->dev, &new->timings);
	if (new->cpu == NULL) {
		dev_del(new->dev);
		gfx_del(new->gfx);
		mem_del(new->mem);
		free(new);
		return NULL;
	}

	return (calc_o)new;
}

void calc_del(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	cpu_del(_calc->cpu);
	dev_del(_calc->dev);
	gfx_del(_calc->gfx);
	mem_del(_calc->mem);

	free(_calc);
}

uint8 calc_read(calc_o calc, unsigned int addr)
{
	_calc_o *_calc = (_calc_o *)calc;

	return mem_read_byte(_calc->mem, addr);
}

void calc_write(calc_o calc, unsigned int addr, uint8 data)
{
	_calc_o *_calc = (_calc_o *)calc;

	mem_write_byte(_calc->mem, addr, data);
}

size_t calc_size(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	return mem_size(_calc->mem);
}

void calc_putimage(calc_o calc, void *data)
{
	_calc_o *_calc = (_calc_o *)calc;

	mem_putimage(_calc->mem, data);
}

unsigned char *calc_getimage(calc_o calc, int *size)
{
	_calc_o *_calc = (_calc_o *)calc;
	unsigned char *ret;

	*size = mem_size(_calc->mem);

	ret = (unsigned char *)malloc(*size);
	if (ret == NULL) {
		_err_report("Unable to allocate space to hold image!");

		return NULL;
	}

	mem_getimage(_calc->mem, ret);

	return ret;
}

void calc_update(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	gfx_update(_calc->gfx);
}

void calc_draw(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	gfx_draw(_calc->gfx);
}

void calc_palette(calc_o calc, int i, uint8 r, uint8 g, uint8 b)
{
	_calc_o *_calc = (_calc_o *)calc;

	gfx_palette(_calc->gfx, i, r, g, b);
}

void calc_keydown(calc_o calc, SDLKey key)
{
	_calc_o *_calc = (_calc_o *)calc;

	dev_keydown(_calc->dev, key);
}

void calc_keyup(calc_o calc, SDLKey key)
{
	_calc_o *_calc = (_calc_o *)calc;

	dev_keyup(_calc->dev, key);
}

int calc_execute(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	return cpu_execute(_calc->cpu);
}

void calc_terminate(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	cpu_terminate(_calc->cpu);
}

int calc_cycle(calc_o calc)
{
	_calc_o *_calc = (_calc_o *)calc;

	return cpu_cycle(_calc->cpu);
}

void calc_interrupt(calc_o calc, unsigned int data)
{
	_calc_o *_calc = (_calc_o *)calc;

	cpu_interrupt(_calc->cpu, data);
}

void calc_report(calc_o calc, FILE *out)
{
	_calc_o *_calc = (_calc_o *)calc;

	cpu_report(_calc->cpu, out);
}

void calc_alter(calc_o calc, const char *regname, unsigned int value)
{
	_calc_o *_calc = (_calc_o *)calc;

	cpu_alter(_calc->cpu, regname, value);
}
